@@ -1,23 +1,24 @@
 #! /usr/bin/env python
-__authors__ = "Dilmurat Yusuf"
+__authors__ = ["Dilmurat Yusuf", "Bert Droesbeke"]
 __copyright__ = "Copyright 2020, Dilmurat Yusuf"
 __email__ = "Dilmurat.yusuf@gmail.com"
 __license__ = "MIT"
 
 import os
 import sys
-import subprocess
-import shlex
 import argparse
 import yaml
 import hashlib
 import ftplib
 import requests
 import uuid
+import numpy as np
+import re
 from genshi.template import TemplateLoader
 from lxml import etree
 import pandas as pd
 import tempfile
+from ena_upload._version import __version__
 
 
 # SettingWithCopyWarning causes false positive
@@ -25,7 +26,7 @@ import tempfile
 pd.options.mode.chained_assignment = None
 
 
-def create_dataframe(schema_tables):
+def create_dataframe(schema_tables, action):
     '''create pandas dataframe from the tables in schema_tables
        and return schema_dataframe
 
@@ -43,8 +44,18 @@ def create_dataframe(schema_tables):
     schema_dataframe = {}
 
     for schema, table in schema_tables.items():
-        df = pd.read_csv(table, sep='\t', comment='#')
-
+        df = pd.read_csv(table, sep='\t', comment='#', dtype = str)
+        # checking for optional columns and if not present adding them
+        if schema == 'sample':
+            optional_columns = ['accession', 'submission_date', 'status', 'scientific_name', 'taxon_id']
+        else:
+            optional_columns = ['accession', 'submission_date', 'status']
+        for header in optional_columns:
+            if not header in df.columns:
+                if header == 'status':
+                    df[header] = action.lower()
+                else:
+                    df[header] = np.nan
         # status column contain action keywords
         # for xml rendering, keywords require uppercase
         # according to scheme definition of submission
@@ -92,6 +103,23 @@ def check_filenames(file_paths, run_df):
         sys.exit(msg)
 
 
+def check_file_checksum(df):
+    '''Return True if 'file_checksum' series contains valid MD5 sums'''
+
+    regex_valid_md5sum = re.compile('^[a-f0-9]{32}$')
+
+    def _is_str_md5sum(x):
+        match = regex_valid_md5sum.fullmatch(x)
+        if match:
+            return True
+        else:
+            return False
+
+    s = df.file_checksum.apply(_is_str_md5sum)
+
+    return s.all()
+
+
 def validate_xml(xsd, xml):
     '''
     validate xml against xsd scheme
@@ -134,7 +162,8 @@ def generate_stream(schema, targets, Template, center, tool):
                                    tool_name=tool['tool_name'],
                                    tool_version=tool['tool_version'])
     else:
-        stream = Template.generate(df=targets, center=center, tool_name=tool['tool_name'],tool_version=tool['tool_version'])
+        stream = Template.generate(
+            df=targets, center=center, tool_name=tool['tool_name'], tool_version=tool['tool_version'])
 
     return stream
 
@@ -152,16 +181,17 @@ def construct_xml(schema, stream, xsd):
 
     validate_xml(xsd, xml_string)
 
-    xml_file = os.path.join(tempfile.gettempdir(), schema + '_' + str(uuid.uuid4()) + '.xml')
+    xml_file = os.path.join(tempfile.gettempdir(),
+                            schema + '_' + str(uuid.uuid4()) + '.xml')
     with open(xml_file, 'w') as fw:
-        fw.write(xml_string.decode("utf-8") )
+        fw.write(xml_string.decode("utf-8"))
 
-    print (f'wrote {xml_file}')
+    print(f'wrote {xml_file}')
 
     return xml_file
 
 
-def actors(template_path, vir):
+def actors(template_path, checklist):
     ''':return: the filenames of schema definitions and templates
     '''
 
@@ -175,25 +205,19 @@ def actors(template_path, vir):
             'submission': 'SRA.submission.xsd',
             'sample': 'SRA.sample.xsd',
             'study': 'SRA.study.xsd'}
-    if vir:
-        templates = {'run': 'ENA_template_runs.xml',
-                    'experiment': 'ENA_template_experiments.xml',
-                    'submission': 'ENA_template_submission.xml',
-                    'sample': 'ENA_template_vir_sample.xml',
-                    'study': 'ENA_template_studies.xml'}
-    else:
-        templates = {'run': 'ENA_template_runs.xml',
-                    'experiment': 'ENA_template_experiments.xml',
-                    'submission': 'ENA_template_submission.xml',
-                    'sample': 'ENA_template_samples.xml',
-                    'study': 'ENA_template_studies.xml'}
+
+    templates = {'run': 'ENA_template_runs.xml',
+                 'experiment': 'ENA_template_experiments.xml',
+                 'submission': 'ENA_template_submission.xml',
+                 'sample': f'ENA_template_samples_{checklist}.xml',
+                 'study': 'ENA_template_studies.xml'}
 
     xsds = add_path(xsds, template_path)
 
     return xsds, templates
 
 
-def run_construct(template_path, schema_targets,  center, vir, tool):
+def run_construct(template_path, schema_targets,  center, checklist, tool):
     '''construct XMLs for schema in schema_targets
 
     :param schema_targets: dictionary of 'schema:targets' generated
@@ -201,12 +225,12 @@ def run_construct(template_path, schema_targets,  center, vir, tool):
     :param loader: object of TemplateLoader in genshi
     :param center: center name used to register ENA Webin
     :param tool: dict of tool_name and tool_version , by default ena-upload-cli
-    :param vir: flag to enable viral sample submission
+    :param checklist: parameter to select a specific sample checklist
 
     :return schema_xmls: dictionary of 'schema:filename'
     '''
 
-    xsds, templates = actors(template_path, vir)
+    xsds, templates = actors(template_path, checklist)
 
     schema_xmls = {}
 
@@ -221,7 +245,7 @@ def run_construct(template_path, schema_targets,  center, vir, tool):
     return schema_xmls
 
 
-def construct_submission(template_path, action, submission_input, center, vir, tool):
+def construct_submission(template_path, action, submission_input, center, checklist, tool):
     '''construct XML for submission
 
     :param action: action for submission -
@@ -231,19 +255,19 @@ def construct_submission(template_path, action, submission_input, center, vir, t
     :param loader: object of TemplateLoader in genshi
     :param center: center name used to register ENA Webin
     :param tool: tool name, by default ena-upload-cli
-    :param vir: flag to enable viral sample submission
+    :param checklist: parameter to select a specific sample checklist
 
     :return submission_xml: filename of submission XML
     '''
 
-    xsds, templates = actors(template_path, vir)
+    xsds, templates = actors(template_path, checklist)
 
     template = templates['submission']
     loader = TemplateLoader(search_path=template_path)
     Template = loader.load(template)
 
     stream = Template.generate(action=action, input=submission_input,
-                               center=center, tool_name=tool['tool_name'],tool_version=tool['tool_version'])
+                               center=center, tool_name=tool['tool_name'], tool_version=tool['tool_version'])
 
     submission_xml = construct_xml('submission', stream, xsds['submission'])
 
@@ -272,23 +296,6 @@ def get_md5(filepath):
 
     return md5sum.hexdigest()
 
-
-def run_cmd(cmd_line):
-    """Execute command line.
-
-    :param cmd_line: the string of command line
-
-    :return output: the string of output from execution
-    """
-
-    args = shlex.split(cmd_line)
-    process = subprocess.Popen(args,
-                     stdout=subprocess.PIPE, 
-                     stderr=subprocess.PIPE)
-    output, stderr = process.communicate()
-    return output
-
-
 def get_taxon_id(scientific_name):
     """Get taxon ID for input scientific_name.
 
@@ -310,6 +317,24 @@ def get_taxon_id(scientific_name):
         msg = f'Oops, no taxon ID avaible for {scientific_name}. Is it a valid scientific name?'
         sys.exit(msg)
 
+def get_scientific_name(taxon_id):
+    """Get scientific name for input taxon_id.
+
+    :param taxon_id: NCBI taxonomy identifier
+    :return scientific_name: scientific name of sample that distinguishes its taxonomy
+    """
+    # endpoint for scientific name
+    url = 'http://www.ebi.ac.uk/ena/taxonomy/rest/tax-id'
+    session = requests.Session()
+    session.trust_env = False
+    r = session.get(f"{url}/{str(taxon_id).strip()}")
+    try:
+        taxon_id = r.json()['scientificName']
+        return taxon_id
+    except ValueError:
+        msg = f'Oops, no scientific name avaible for {taxon_id}. Is it a valid taxon_id?'
+        sys.exit(msg)
+
 
 def submit_data(file_paths, password, webin_id):
     """Submit data to webin ftp server.
@@ -319,20 +344,20 @@ def submit_data(file_paths, password, webin_id):
     """
 
     try:
-        print ("\nconnecting to ftp.webin.ebi.ac.uk....")
+        print("\nconnecting to ftp.webin.ebi.ac.uk....")
         ftp = ftplib.FTP("webin.ebi.ac.uk", webin_id, password)
     except IOError:
-        print (ftp.lastErrorText())
-        print ("ERROR: could not connect to the ftp server.\
+        print(ftp.lastErrorText())
+        print("ERROR: could not connect to the ftp server.\
                Please check your login details.")
 
     for filename, path in file_paths.items():
-        print (f'uploading {path}')
+        print(f'uploading {path}')
         ftp.storbinary(f'STOR {filename}', open(path, 'rb'))
         msg = ftp.storbinary(f'STOR {filename}', open(path, 'rb'))
-        print (msg)
+        print(msg)
 
-    print (ftp.quit())
+    print(ftp.quit())
 
 def columns_to_update(df):
     '''
@@ -363,13 +388,13 @@ def send_schemas(schema_xmls, url, webin_id, password):
     password -- ENA password of user
     '''
 
-
-    sources = [(schema.upper(), (source, open(source, 'rb'))) for schema, source in schema_xmls.items()]
+    sources = [(schema.upper(), (source, open(source, 'rb')))
+               for schema, source in schema_xmls.items()]
     session = requests.Session()
     session.trust_env = False
     r = session.post(f"{url}",
-                    auth=(webin_id,password),
-                    files=sources)
+                     auth=(webin_id, password),
+                     files=sources)
 
     print(r.text)
     return r
@@ -387,7 +412,7 @@ def process_receipt(receipt, action):
                                    'accession', 'submission_date'
     '''
     receipt_root = etree.fromstring(receipt)
-    
+
     success = receipt_root.get('success')
 
     if success == 'true':
@@ -408,7 +433,8 @@ def process_receipt(receipt, action):
         update_list = []
         print(f"\n{ena_type.capitalize()} accession details:")
         for element in update:
-            extract = (element.get('alias'), element.get('accession'), receiptDate, status[action])
+            extract = (element.get('alias'), element.get(
+                'accession'), receiptDate, status[action])
             print("\t".join(extract))
             update_list.append(extract)
         # used for labelling dataframe
@@ -432,7 +458,8 @@ def process_receipt(receipt, action):
         schema_update['sample'] = make_update(sample_update, 'sample')
 
     if experiment_update:
-        schema_update['experiment'] = make_update(experiment_update, 'experiment')
+        schema_update['experiment'] = make_update(
+            experiment_update, 'experiment')
 
     if run_update:
         schema_update['run'] = make_update(run_update, 'run')
@@ -478,7 +505,7 @@ def update_table(schema_dataframe, schema_targets, schema_update):
             if schema == 'sample':
                 dataframe.loc[index,
                               'taxon_id'] = targets.loc[index, 'taxon_id']
-            if schema == 'run':
+            elif schema == 'run':
                 # since index is set to alias
                 # then targets of run can have multiple rows with
                 # identical index because each row has a file
@@ -503,7 +530,7 @@ def save_update(schema_tables_, schema_dataframe_):
         :dataframe: a dataframe
     """
 
-    print ('\nSaving updates in new tsv tables::')
+    print('\nSaving updates in new tsv tables::')
     for schema in schema_tables_:
         table = schema_tables_[schema]
         dataframe = schema_dataframe_[schema]
@@ -511,7 +538,7 @@ def save_update(schema_tables_, schema_dataframe_):
         file_name, file_extension = os.path.splitext(table)
         update_name = f'{file_name}_updated{file_extension}'
         dataframe.to_csv(update_name, sep='\t')
-        print (f'save updates in {update_name}')
+        print(f'save updates in {update_name}')
 
 
 class SmartFormatter(argparse.HelpFormatter):
@@ -541,7 +568,9 @@ def process_args():
                                      corresponding the ENA objects -- STUDY,
                                      SAMPLE, EXPERIMENT and RUN.''',
                                      formatter_class=SmartFormatter)
-    parser.add_argument('--version', action='version', version='%(prog)s 0.3.0')
+    parser.add_argument('--version', action='version',
+                        version='%(prog)s '+__version__)
+
     parser.add_argument('--action',
                         choices=['add', 'modify', 'cancel', 'release'],
                         required=True,
@@ -564,37 +593,45 @@ def process_args():
 
     parser.add_argument('--data',
                         nargs='*',
-                        help='Data for submission',
+                        help='data for submission',
                         metavar='FILE')
 
     parser.add_argument('--center',
                         dest='center_name',
                         required=True,
                         help='specific to your Webin account')
+
+    parser.add_argument('--checklist', help="specify the sample checklist with following pattern: ERC0000XX, Default: ERC000011", dest='checklist',
+                        default='ERC000011')
+
     parser.add_argument('--tool',
                         dest='tool_name',
                         default='ena-upload-cli',
-                        help='Specify the name of the tool this submission is done with. Default: ena-upload-cli')
+                        help='specify the name of the tool this submission is done with. Default: ena-upload-cli')
 
     parser.add_argument('--tool_version',
                         dest='tool_version',
-                        default='0.3.0',
-                        help='Specify the version of the tool this submission is done with.')
+                        default=__version__,
+                        help='specify the version of the tool this submission is done with')
+
+    parser.add_argument('--no_upload',
+                        default=False,
+                        action="store_true",
+                        help='Indicate if no upload should be performed and you like to submit a RUN object (e.g. if uploaded was done separately).')
 
     parser.add_argument('--secret',
                         required=True,
                         help='.secret.yml file containing the password and Webin ID of your ENA account')
 
-    parser.add_argument('-d', '--dev', help="Flag to use the dev/sandbox endpoint of ENA.", action="store_true")
-    
-    parser.add_argument('--vir', help="Flag to use the viral sample template.", action="store_true")
+    parser.add_argument(
+        '-d', '--dev', help="flag to use the dev/sandbox endpoint of ENA", action="store_true")
 
     args = parser.parse_args()
 
     # check if any table is given
     tables = set([args.study, args.sample, args.experiment, args.run])
     if tables == {None}:
-        parser.error('requires at least one table for submission')
+        parser.error('Requires at least one table for submission')
 
     # check if .secret file exists
     if args.secret:
@@ -602,7 +639,6 @@ def process_args():
             msg = f"Oops, the file {args.secret} does not exist"
             parser.error(msg)
 
-
     # check if data is given when adding a 'run' table
     if args.action == 'add' and args.run is not None:
         if args.data is None:
@@ -634,23 +670,24 @@ def collect_tables(args):
     return schema_tables
 
 
-def main ():
+def main():
     args = process_args()
     action = args.action.upper()
     center = args.center_name
-    tool = {'tool_name':args.tool_name, 'tool_version':args.tool_version}
+    tool = {'tool_name': args.tool_name, 'tool_version': args.tool_version}
     dev = args.dev
-    vir = args.vir
+    checklist = args.checklist
     secret = args.secret
 
-    secret_file = open(secret, "r")
-    credentials = yaml.load(secret_file, Loader=yaml.FullLoader)
+    with open(secret, 'r') as secret_file:
+        credentials = yaml.load(secret_file, Loader=yaml.FullLoader)
 
     password = credentials['password'].strip()
     webin_id = credentials['username'].strip()
 
     if not password or not webin_id:
-        print( f"Oops, file {args.secret} does not contain a password or username")
+        print(
+            f"Oops, file {args.secret} does not contain a password or username")
     secret_file.close()
 
     # ? a function needed to convert characters e.g. # -> %23 in password
@@ -659,7 +696,7 @@ def main ():
     schema_tables = collect_tables(args)
 
     # create dataframe from table
-    schema_dataframe = create_dataframe(schema_tables)
+    schema_dataframe = create_dataframe(schema_tables, action)
 
     # ? add a function to sanitize characters
     # ? print 'validate table for specific action'
@@ -676,49 +713,66 @@ def main ():
         if 'run' in schema_targets:
             # a dictionary of filename:file_path
             # ? do I have to define the absolute path
-            file_paths = {os.path.basename(path): os.path.abspath(path)
-                          for path in args.data}
-
             df = schema_targets['run']
 
-            # check if file names identical between command line and table
-            # if not, system exits
-            check_filenames(file_paths, df)
+            if args.no_upload:
+                file_paths = {}
+                print("No files will be uploaded, remove `--no_upload' argument to perform upload.")
+            else:  # check supplied datafiles only if doing upload
+                file_paths = {os.path.basename(path): os.path.abspath(path)
+                                                   for path in args.data}
+                # check if file names identical between command line and table
+                # if not, system exits
+                check_filenames(file_paths, df)
 
-            file_md5 = {filename: get_md5(path) for filename, path
-                        in file_paths.items()}
+            # generate MD5 sum if not supplied in table
+            if not check_file_checksum(df):
+                print("No valid checksums found, generate now...", end=" ")
+                file_md5 = {filename: get_md5(path) for filename, path
+                            in file_paths.items()}
+
+                # update schema_targets wih md5 hash
+                md5 = df['file_name'].apply(lambda x: file_md5[x]).values
+                df.loc[:, 'file_checksum'] = md5
+                print("done.")
 
-            # update schema_targets wih md5 hash
-            md5 = df['file_name'].apply(lambda x: file_md5[x]).values
-            df.loc[:, 'file_checksum'] = md5
             schema_targets['run'] = df
 
             # submit data to webin ftp server
-            submit_data(file_paths, password, webin_id)
+            if not args.no_upload:
+                submit_data(file_paths, password, webin_id)
 
         # when adding sample
-        # update schema_targets with taxon ids
+        # update schema_targets with taxon ids or scientific names
         if 'sample' in schema_targets:
             df = schema_targets['sample']
-
-            # retrieve taxon id using scientific name
-            print ('retrieving taxon IDs...')
-            taxonID = df['scientific_name'].apply(get_taxon_id).values
-            print ('taxon IDs are retrieved')
-            df.loc[:, 'taxon_id'] = taxonID
+            print('retrieving taxon IDs and scientific names if needed')
+            for index, row in df.iterrows():
+                if pd.notna(row['scientific_name']) and pd.isna(row['taxon_id']):
+                    # retrieve taxon id using scientific name
+                    taxonID = get_taxon_id(row['scientific_name'])
+                    df.loc[index, 'taxon_id'] = taxonID
+                elif pd.notna(row['taxon_id']) and pd.isna(row['scientific_name']):
+                    # retrieve scientific name using taxon id
+                    scientificName = get_scientific_name(row['taxon_id'])
+                    df.loc[index, 'scientific_name'] = scientificName
+                elif pd.isna(row['taxon_id']) and pd.isna(row['scientific_name']):
+                    sys.exit(f"No taxon_id or scientific_name was given with sample {row['alias']}.")
+            print('taxon IDs and scientific names are retrieved')
             schema_targets['sample'] = df
 
     # ? need to add a place holder for setting up
     base_path = os.path.abspath(os.path.dirname(__file__))
-    template_path = os.path.join(base_path,'templates')
+    template_path = os.path.join(base_path, 'templates')
     if action in ['ADD', 'MODIFY']:
         # when ADD/MODIFY,
         # requires source XMLs for 'run', 'experiment', 'sample', 'experiment'
         # schema_xmls record XMLs for all these schema and following 'submission'
-        schema_xmls = run_construct(template_path, schema_targets, center, vir, tool)
+        schema_xmls = run_construct(
+            template_path, schema_targets, center, checklist, tool)
 
         submission_xml = construct_submission(template_path, action,
-                                              schema_xmls, center, vir, tool)
+                                              schema_xmls, center, checklist, tool)
 
     elif action in ['CANCEL', 'RELEASE']:
         # when CANCEL/RELEASE, only accessions needed
@@ -726,7 +780,7 @@ def main ():
         schema_xmls = {}
 
         submission_xml = construct_submission(template_path, action,
-                                              schema_targets, center, vir, tool)
+                                              schema_targets, center, checklist, tool)
 
     schema_xmls['submission'] = submission_xml
 
@@ -753,20 +807,6 @@ def main ():
     # save updates in new tables
     save_update(schema_tables, schema_dataframe)
 
-    # print 'optional: need a testing script'
-
-    # receipt example for cancel
-    # <?xml version="1.0" encoding="UTF-8"?>
-    # <?xml-stylesheet type="text/xsl" href="receipt.xsl"?>
-    # <RECEIPT receiptDate="2017-12-05T17:07:13.874Z" submissionFile="submission_cancel.xml" success="true">
-    #      <MESSAGES>
-    #           <INFO>Submission has been committed.</INFO>
-    #           <INFO>This submission is a TEST submission and will be discarded within 24 hours</INFO>
-    #      </MESSAGES>
-    #      <ACTIONS/>
-    #      <ACTIONS/>
-    # </RECEIPT>
-
 
 if __name__ == "__main__":
 
